
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdint.h>
#include <pthread.h>

#include "poolmemalloc.h"

#define POINTERSIZE (sizeof(void*))

#define ALIGN_SIZE 8
#define MAX_ALLOC_DIVIDE 2

#define ALIGN_UP(value,alignment) (((uintptr_t)value + alignment - 1) & -alignment)

#define ALIGN_UP_ALIGN(p) ALIGN_UP(p,ALIGN_SIZE)
#define ALIGN_UP_POINTER(p) ALIGN_UP(p,POINTERSIZE)

#define CHAR(a) ((char*)a)

#if 0
   static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
#  define LOCK() pthread_mutex_lock(&lock)
#  define UNLOCK() pthread_mutex_unlock(&lock)
#else
#  define LOCK() 
#  define UNLOCK() 
#endif


ST poolmem_t *poolmem_create(size_t size,void *mem,int max_mem_size){
  poolmem_t *poolmem=malloc(sizeof(poolmem_t));

  poolmem->mem=mem==NULL?my_calloc(1,size):memset(mem,0,size);
  if(poolmem->mem==NULL){
    free(poolmem);
    fprintf(stderr,"create_poolmem, no memory\n");
    abort();
    return NULL;
  }

  poolmem->mem_free=poolmem->mem;
  poolmem->mem_free_end=poolmem->mem_free + size;
  poolmem->free=my_calloc(sizeof(poolmem_element_t*),max_mem_size);
  if(poolmem->free==NULL){
    fprintf(stderr,"create_poolmem (2), no memory\n");
    abort();
    if(mem==NULL)
      free(poolmem->mem);
    free(poolmem);
    return NULL;
  }

  poolmem->max_alloc=size/MAX_ALLOC_DIVIDE;
  return poolmem;
}

ST void poolmem_delete(poolmem_t *poolmem){
  free(poolmem->mem);
  free(poolmem->free);
  free(poolmem);
}



// Approx. 12 instructions worst-case
//
// NOTE! size must be aligned up to POINTERSIZE before calling.
//
ST void *poolmem_alloc(poolmem_t *poolmem,size_t size){
  poolmem_element_t *element=NULL;

  assert(size==ALIGN_UP_POINTER(size));

  if(size > poolmem->max_alloc){
    return NULL;
  }

  LOCK();

  bytes_used+=size;
  if(bytes_used>max_bytes_used)
    max_bytes_used=bytes_used;

  {
    element=poolmem->free[size/POINTERSIZE];

    if(element!=NULL){
      poolmem->free[size/POINTERSIZE] = element->next;
      //fprintf(stderr,"  RHEPP0 %d (size: %d)\n",((char*)element)-((char*)poolmem->mem),size);
      element->next=NULL;
      goto exit;
    }
    
    element = (poolmem_element_t*) poolmem->mem_free;
    poolmem->mem_free += size;
    
    if(poolmem->mem_free > poolmem->mem_free_end){
      poolmem->mem_free -= size;
      element=NULL;
      goto exit;
    }
    
    //fprintf(stderr,"   POOLMEM ALLOC %d (size: %d) Element: %p\n",((char*)element)-((char*)poolmem->mem),size,element);
  }

 exit:
  UNLOCK();
#ifdef DEBUG_DOUBLEFREE	  
  if(element!=NULL)
    ((char*)element)[16]='u';
#endif
  return element;
}

// Approx max 8 instructions.
//
ST void poolmem_free(poolmem_t *poolmem,void *v_start,size_t size){
  char *start=v_start;
  char *end=v_start+size;

  //fprintf(stderr,"   POOLMEM FREE. Size: %d, pos: %p\n",size,start);
#ifdef DEBUG_DOUBLEFREE
  //fprintf(stderr,"start[16]: -%c- (%d)\n",start[16],(int)start[16]);
  assert(start[16]!='a');
  assert(start[16]=='u');
  start[16]='a';
#endif

  LOCK();

  bytes_used-=size;

  if(end==poolmem->mem_free){
    poolmem->mem_free=start;
  }else{
    poolmem_element_t *element=v_start;
    element->next=poolmem->free[size/POINTERSIZE];
    poolmem->free[size/POINTERSIZE]=element;
  }

  UNLOCK();
}

#ifdef TESTING
int main(){
  poolmem_t *pm=create_poolmem(1024*104,NULL);
  void *ai=poolmem_alloc(pm,127);
  poolmem_free(pm,ai);
  void *ai2=poolmem_alloc(pm,126);
  printf("size: %d %p %p %p, size:%d\n",sizeof(poolmem_element_pre_t),pm->mem,ai,ai2,poolmem_getsize(ai2));
  return 0;
}
#endif
